import asyncio
import collections
import inspect
import logging

from aiozk import WatchEvent, exc

from .recipe import Recipe

log = logging.getLogger(__name__)


def maybe_future(fut, loop):
    if inspect.isawaitable(fut):
        asyncio.ensure_future(fut, loop=loop)


class BaseWatcher(Recipe):

    watched_events = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.callbacks = collections.defaultdict(set)
        self.loops = {}

    def add_callback(self, path, callback):
        self.callbacks[path].add(callback)

        if path not in self.loops:
            self.loops[path] = asyncio.ensure_future(self.watch_loop(path), loop=self.client.loop)

    def remove_callback(self, path, callback):
        self.callbacks[path].discard(callback)

        if not self.callbacks[path]:
            self.callbacks.pop(path)
            self.loops.pop(path).cancel()

    async def fetch(self, path):
        raise NotImplementedError

    async def watch_loop(self, path):
        while self.callbacks[path]:
            log.debug("Fetching data for %s", path)
            try:
                result = await self.fetch(path)
            except exc.NoNode:
                result = exc.NoNode
            except exc.ZKError as e:
                log.exception('Exception in watch loop: {}'.format(e))
                log.info('Waiting for safe state...')
                await self.client.session.ensure_safe_state()
                continue
            except Exception:
                log.exception('Not handled in watch loop:')
                raise

            for callback in self.callbacks[path].copy():
                maybe_future(callback(result), loop=self.client.loop)
            if WatchEvent.CREATED not in self.watched_events and result == exc.NoNode:
                return
            try:
                await self.client.wait_for_events(self.watched_events, path)
            except asyncio.CancelledError:
                pass
            except Exception as e:
                log.exception('Not handled in wait_for_events:')
                print('Not handled: {!r}'.format(e))
                raise



import asyncio
import logging
import time

from aiozk import exc, states

from .sequential import SequentialRecipe

log = logging.getLogger(__name__)


class BaseLock(SequentialRecipe):
    async def wait_in_line(self, znode_label, timeout=None, blocked_by=None):
        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        await self.create_unique_znode(znode_label)

        while True:
            if time_limit and time.time() >= time_limit:
                await self.delete_unique_znode(znode_label)
                raise exc.TimeoutError

            owned_positions, contenders = await self.analyze_siblings()
            if znode_label not in owned_positions:
                raise exc.SessionLost

            blockers = contenders[:owned_positions[znode_label]]
            if blocked_by:
                blockers = [
                    contender for contender in blockers
                    if self.determine_znode_label(contender) in blocked_by
                ]

            if not blockers:
                break

            try:
                await self.wait_on_sibling(blockers[-1], timeout)
            except exc.TimeoutError:
                await self.delete_unique_znode(znode_label)

        return self.make_contextmanager(znode_label)

    def make_contextmanager(self, znode_label):
        state = {"acquired": True}

        def still_acquired():
            return state["acquired"]

        state_fut = self.client.session.state.wait_for(states.States.LOST)

        async def handle_session_loss():
            await state_fut
            if not state["acquired"]:
                return

            log.warning(
                "Session expired at some point, lock %s no longer acquired.",
                self)
            state["acquired"] = False

        fut = asyncio.ensure_future(handle_session_loss(),
                                    loop=self.client.loop)

        async def on_exit():
            state["acquired"] = False
            if not fut.done():
                if not state_fut.done():
                    self.client.session.state.remove_waiting(
                        state_fut, states.States.LOST)
                fut.cancel()

            await self.delete_unique_znode(znode_label)

        class Lock:
            async def __aenter__(self):
                return self

            async def __aexit__(self, exc_type, exc, tb):
                await on_exit()

        return Lock()


class LockLostError(exc.ZKError):
    pass



import asyncio
import logging

from aiozk import exc, WatchEvent

from .sequential import SequentialRecipe

log = logging.getLogger(__name__)


class DoubleBarrier(SequentialRecipe):

    def __init__(self, base_path, min_participants):
        super().__init__(base_path)
        self.min_participants = min_participants

    @property
    def sentinel_path(self):
        return self.sibling_path("sentinel")

    async def enter(self, timeout=None):
        log.debug("Entering double barrier %s", self.base_path)
        barrier_lifted = self.client.wait_for_events(
            [WatchEvent.CREATED], self.sentinel_path
        )

        exists = await self.client.exists(path=self.sentinel_path, watch=True)

        await self.create_unique_znode("worker")

        _, participants = await self.analyze_siblings()

        if exists:
            return

        elif len(participants) >= self.min_participants:
            await self.create_znode(self.sentinel_path)
            return

        try:
            if timeout:
                await asyncio.wait_for(barrier_lifted, timeout)
            else:
                await barrier_lifted
        except asyncio.TimeoutError:
            raise exc.TimeoutError

    async def leave(self, timeout=None):
        log.debug("Leaving double barrier %s", self.base_path)
        owned_positions, participants = await self.analyze_siblings()
        while len(participants) > 1:
            if owned_positions["worker"] == 0:
                await self.wait_on_sibling(participants[-1], timeout)
            else:
                await self.delete_unique_znode("worker")
                await self.wait_on_sibling(participants[0], timeout)

            owned_positions, participants = await self.analyze_siblings()

        if len(participants) == 1 and "worker" in owned_positions:
            await self.delete_unique_znode("worker")
            try:
                await self.client.delete(self.sentinel_path)
            except exc.NoNode:
                pass



import asyncio
import pytest
from aiozk import states
from aiozk.exc import TimeoutError


@pytest.mark.asyncio
async def test_shared_lock(zk, path):

    shared = zk.recipes.SharedLock(path)
    got_lock = False
    async with await shared.acquire_write():
        got_lock = True

    assert got_lock

    got_released = False
    async with await shared.acquire_write():
        got_released = True

    assert got_released
    await zk.delete(path)


@pytest.mark.asyncio
async def test_shared_lock_timeout(zk, path):
    got_lock = False
    async with await zk.recipes.SharedLock(path).acquire_write():
        got_lock = True

        with pytest.raises(TimeoutError):
            await zk.recipes.SharedLock(path).acquire_write(timeout=1)

    assert got_lock

    got_released = False
    async with await zk.recipes.SharedLock(path).acquire_write():
        got_released = True

    assert got_released

    await zk.deleteall(path)


@pytest.mark.asyncio
async def test_delete_unique_znode_on_timeout(zk, path):
    lock = zk.recipes.SharedLock(path)
    try:
        # timeout -1 here to force a timeout error
        await lock.wait_in_line('write', timeout=-1)
    except TimeoutError:
        pass

    _, contenders = await lock.analyze_siblings()

    # when we have a timeout error the contender must be deleted.
    assert not contenders

    await zk.delete(path)


def inspect_waiting_loss_handlers(zk, tag):
    waitings = zk.session.state.waitings(
        states.States.LOST)[states.States.LOST]
    tasks = [task for task in asyncio.Task.all_tasks() if not task.done()]

    print(f'({tag}) waitings={waitings}, len={len(waitings)}')
    print(f'({tag}) tasks={tasks} len={len(tasks)}')

    return waitings, tasks


@pytest.mark.asyncio
async def test_remove_session_loss_handler_after_lock_released(zk, path):
    waitings_orig, tasks_orig = inspect_waiting_loss_handlers(zk, 'original')
    lock = zk.recipes.SharedLock(path)
    async with await lock.acquire_write():
        waitings_locking, tasks_locking = inspect_waiting_loss_handlers(
            zk, 'locking')

    waitings_released, tasks_released = inspect_waiting_loss_handlers(
        zk, 'released')

    assert len(waitings_locking) > len(waitings_released)
    assert len(waitings_orig) == len(waitings_released)
    assert len(tasks_locking) > len(tasks_released)
    assert len(tasks_orig) == len(tasks_released)

    await zk.delete(path)
